#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.algo - Arbre et Trie (correction)
# 
# Correction.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Exercice 1
# 

# In[2]:


import random
def mot_alea (l) :
    l = [ chr(97+random.randint(0,25)) for i in range(l) ]
    return "".join(l)

taille  = 20
N       = 10000
mots    = [ mot_alea(taille) for _ in range (N) ]
print(len(mots))


# ## Exercice 2

# In[3]:


import time
debut = time.perf_counter()
for k in mots :
    i = mots.index(k)
fin = time.perf_counter()
print ("recherche simple",fin - debut)


# Avec ``%timeit`` :

# In[4]:


get_ipython().run_line_magic('timeit', 'for k in mots : i = mots.index(k)')


# ## Exercice 3 : recherche dichotomique

# In[5]:


def dicho (mots, x) :
    a = 0
    b = len(mots)-1
    while a < b :
        m = (a+b)//2
        t = mots[m] 
        if t < x : 
            b = m-1 
        elif t == x :
            return m
        else :
            a = m+1
    return a

mots.sort()

debut = time.perf_counter()
for k in mots :
    i = dicho(mots, k)
fin = time.perf_counter()
print ("dichotomie",fin - debut)


# In[6]:


get_ipython().run_line_magic('timeit', 'for k in mots : i = dicho(mots, k)')


# ## Exercice 4

# In[7]:


import math

for N in [10, 100, 1000, 10000, 100000] :
    mots    = [ mot_alea(taille) for _ in range (N) ]
    tolook  = [ mots[ random.randint(0,len(mots)-1) ] for i in range(0,1000) ]
    mots.sort()
    
    debut = time.perf_counter()
    for k in tolook :
        i = mots.index(k)
    fin = time.perf_counter()    
    ds = fin-debut
    
    debut = time.perf_counter()
    for k in tolook :
        i = dicho(mots, k)
    fin = time.perf_counter()    
    dd = fin-debut
    
    print(N, "simple",ds, "dicho",dd, "ratio", ds / max(dd, 1), " ratio théorique ",
          len(mots)/math.log(len(mots)) * math.log(2)/30)


# In[8]:


for N in [10, 100, 1000, 10000, 100000] :
    print("N=",N)
    mots    = [ mot_alea(taille) for _ in range (N) ]
    tolook  = [ mots[ random.randint(0,len(mots)-1) ] for i in range(0,1000) ]
    mots.sort()
    
    get_ipython().run_line_magic('timeit', 'for k in tolook : i = mots.index(k)')
    get_ipython().run_line_magic('timeit', 'for k in tolook : i = dicho(mots, k)')


# Soit $N$ le nombre de mots dans la liste :
# 
# * coût de la recherche simple : $O(N)$ 
# * coût de la recherche dichotomique : $O(\ln N)$ 
# 
# Le ratio $N/\ln N$ mesuré en pratique devrait être sensiblement égal au ratio théorique à un facteur multiplication près. Le trie du tableau qui précède la recherche dichotomique n'est pas pris en compte. Plus on effectue de recherche, plus son coût devient marginal. Ce coût explique néanmoins pourquoi on ne fait pas toujours une recherche dichotomique.

# ## Exercice 5 : trie

# In[9]:


def build_trie(mots) :
    trie = { }
    for m in mots :
        r = trie
        for c in m :
            if c not in r : r[c] = { }
            r = r[c]
    return trie

mots = [ "aaa", "aba", "aab", "baa", "bbb", "bba", "bab" ]

trie = build_trie(mots)
print(trie)


# Quelques explications supplémentaires concernant cette correction :
# 
# * [Question à propos de trie](http://www.xavierdupre.fr/app/ensae_teaching_cs/helpsphinx/questions/question_2014.html#question-a-propos-de-trie)

# ## Exercice 6 : recherche dans un trie

# In[10]:


def lookup(trie, m) :
    r = trie
    for c in m :
        if c in r :
            r = r[c]
        else :
            return False
    return True

for k in mots :
    print(k, lookup(trie, k))
print("bcc", lookup(trie, "bcc"))    


# Pour illustrer la structure d'arbre du trie, on l'affiche avec la fonction suivante : 

# In[11]:


def print_trie(trie, niveau = 0):
    for k,v in sorted(trie.items()):
        print("  " * niveau + k)
        if len(v) > 0 :
            print_trie(v, niveau+1)
print_trie(trie)


# Il reste un inconvénient à cette représentation. Si on construire le trie pour le mot ``["aaa"]`` ou les mots ``["aa","aaa]"``, on obtient le même trie :

# In[12]:


print_trie (build_trie( ["aaa"]) )


# In[13]:


print_trie (build_trie( ["aaa", "aa"]) )


# Pour éviter cela, le plus simple est de représenter la fin d'un mot comme un caractère à part entière.

# In[14]:


print_trie (build_trie( ["aaa*"]) )


# In[15]:


print_trie (build_trie( ["aaa*", "aa*"]) )


# ## Exercice 7
# 
# Soit $L$ la longueur maximale des mots et $C$ le nombre de lettres distinctes, avec un trie, le coût de la recherche est majoré par : $O(L \ln C)$. On reprend le code de l'exercice 5 et on ajoute le code associé au trie. On effectue 10000 recherches au lieu de 1000 pour avoir une meilleure estimation de la différence (pour vous en convaincre, il suffit comparer les temps obtenus par deux exécution de ce même code).

# In[16]:


for N in [10, 100, 1000, 10000, 100000, 200000, 400000] :
    mots    = [ mot_alea(taille) for _ in range (N) ]
    tolook  = [ mots[ random.randint(0,len(mots)-1) ] for i in range(0,10000) ]
    trie    = build_trie(mots)
    mots.sort()
    
    debut = time.perf_counter()
    for k in tolook :
        i = dicho(mots, k)
    fin = time.perf_counter()    
    dd = fin-debut
    
    debut = time.perf_counter()
    for k in tolook :
        i = lookup(trie, k)
    fin = time.perf_counter()
    dt = fin - debut

    print(N, "dicho",dd, "trie", dt)


# In[17]:


for N in [10, 100, 1000, 10000, 100000, 200000, 400000] :
    print("N=",N)
    mots    = [ mot_alea(taille) for _ in range (N) ]
    tolook  = [ mots[ random.randint(0,len(mots)-1) ] for i in range(0,10000) ]
    trie    = build_trie(mots)
    mots.sort()
    
    get_ipython().run_line_magic('timeit', 'for k in tolook : i = dicho(mots, k)')
    get_ipython().run_line_magic('timeit', 'for k in tolook : i = lookup(trie, k)')


# Encore une fois, le temps de construction du trie n'est pas pris en compte. Plus il y a de recherche à effectuer, plus il devient négligeable.
# 
# Le dictionnaire est un object courant dans la plupart des languages. En python, celui-ci utilise une [table de hachage](http://fr.wikipedia.org/wiki/Table_de_hachage) et le coût d'accès à un élément n'est pas en $O(\ln n)$ mais en $O(n)$ (voir [time complexity](https://wiki.python.org/moin/TimeComplexity)). En C++, le dictionnaire (ou [map](http://www.cplusplus.com/reference/map/map/)) utilise un arbre binaire et l'accès à un élément a un coût logarithmique : [Standard C++ Containers](http://www.cs.northwestern.edu/~riesbeck/programming/c++/stl-summary.html).

# ## Plus en détails

# La recherche dichotomique est équivalente à celle opérée avec un [arbre binaire de recherche](http://fr.wikipedia.org/wiki/Arbre_binaire_de_recherche) (si ce dernier est [équilibré](http://fr.wikipedia.org/wiki/Arbre_B) ou [arbre rouge/noir](http://fr.wikipedia.org/wiki/Arbre_bicolore)). Ce dernier consiste à classer les éléments par ordre alphabétique. Un arbre est souvent représenté par une classe et non par un dictionnaire comme la dernière partie de cette session le laissait supposer.

# In[18]:


class Arbre:
    def __init__(self, value):
        self.value = value
        self.children = [ ]
        
    def add_child(self, child):
        self.children.append(child)
        
    def __str__(self):
        rows = [ "value={0}".format(self.value) ]
        for c in self.children:
            s = str(c)
            lines = [ "    " + l for l in s.split("\n") ]
            rows.extend(lines)
        return "\n".join(rows)
    
root = Arbre("racine")
child1 = Arbre("child 1")
child1.add_child ( Arbre("child 2") )
child1.add_child ( Arbre("child 1000") )
root.add_child(child1)
root.add_child( Arbre ("child 3") )
print(root)     


# Les arbres sont des graphes particuliers car il ne contiennent pas de cycles. Il est possible de parcourir les noeuds, de les numéroter. Ils sont très utilisés en machine learning avec les [arbres de décisions](http://fr.wikipedia.org/wiki/Arbre_de_d%C3%A9cision) ou les [random forests](http://en.wikipedia.org/wiki/Random_forest). Ils sont parfois cachés comme dans le cas de la recherche dichotomique qui peut être implémentée à partir d'une structure d'arbre.
# 
# Dans le cas de le recherche dichotomique, on suppose que le nombre de noeuds fils est toujours 2. L'ordre alphabétique est le suivant : noeuds fils 1, noeud courant, noeud fils 2. Les deux noeuds fils pourraient être nuls. L'implémentation de l'arbre serait la suivante :

# In[19]:


class ArbreDicho:
    def __init__(self, value):
        self.value = value
        self.before = None
        self.after = None
        
    def __str__(self):
        return "value={0}".format(self.value)
        
    def add_before(self, child):
        self.before = child

    def add_after(self, child):
        self.after = child
        
    def find(self, word):
        if self.value == word : return self
        elif word < self.value : 
            if self.before is None : return None
            else : return self.before.find(word)
        else : 
            if self.after is None : return None
            else : return self.after.find(word)
            
    def sorted_list(self):
        res = [ ]
        if self.before is not None: res.extend ( self.before.sorted_list() )
        res.append(self.value)
        if self.after is not None: res.extend ( self.after.sorted_list() )
        return res

# on crée un arbre dont les noeuds vérifient la propriété énoncé plus haut (les mots apparaissent dans le bon ordre)
root = ArbreDicho("milieu")
root.add_before(ArbreDicho("avant"))
root.add_after(ArbreDicho("zillion"))
root.before.add_before(ArbreDicho("alphabet"))
root.before.add_after(ArbreDicho("avant après"))

# on vérifie que c'est bien le cas
all = root.sorted_list()
assert all == sorted(all)
print(all)

# on effectue la recherche
for a in all:
    f = root.find(a)
    print(f)


# In[20]: